/*******************************************************************************
 * Copyright (c) 2014-2016 Zeligsoft (2009) Limited  and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/

package org.eclipse.papyrusrt.codegen.cpp.statemachines.flat

import java.util.ArrayList
import org.eclipse.emf.ecore.EObject
import org.eclipse.papyrusrt.codegen.CodeGenPlugin
import org.eclipse.papyrusrt.codegen.UserEditableRegion
import org.eclipse.papyrusrt.codegen.UserEditableRegion.Label
import org.eclipse.papyrusrt.codegen.statemachines.transformations.FlatteningTransformer
import org.eclipse.papyrusrt.xtumlrt.common.ActionCode
import org.eclipse.papyrusrt.xtumlrt.common.CommonElement
import org.eclipse.papyrusrt.xtumlrt.common.Entity
import org.eclipse.papyrusrt.xtumlrt.common.NamedElement
import org.eclipse.papyrusrt.xtumlrt.statemach.State
import org.eclipse.papyrusrt.xtumlrt.statemach.Transition
import org.eclipse.papyrusrt.xtumlrt.statemachext.GuardAction
import org.eclipse.papyrusrt.xtumlrt.statemachext.TransitionAction
import org.eclipse.papyrusrt.xtumlrt.trans.from.uml.UML2xtumlrtTranslator
import org.eclipse.papyrusrt.xtumlrt.umlrt.RTTrigger
import org.eclipse.papyrusrt.xtumlrt.util.QualifiedNames
import org.eclipse.papyrusrt.xtumlrt.statemach.StateMachine

/**
 * Utility class for generating user editable region tag.
 * 
 * @author Young-Soo Roh
 */
class UserCodeTagUtil {

	/**
	 * Get information required to calculate user edit region tag
	 */
	static def Label generateLabel(ActionCode context, UML2xtumlrtTranslator translator, FlatteningTransformer flattener, Entity capsuleOrClass) {
		val label = new Label
		if (capsuleOrClass.eResource != null) {
			label.uri = capsuleOrClass.eResource.URI.toString
		} else {
			var src = translator.getSource(context)
			if(src == null){
				src = translator.getSource(context.eContainer as CommonElement)
			}
			if (src !== null){
				label.uri = src.eResource.URI.toString
			}
		}
		val originalOwner = flattener.getOriginalOwner(context as CommonElement)
		label.qualifiedName = originalOwner.getQualifiedName(capsuleOrClass)
		label.type = context.eClass.name.replace("Action", "").toLowerCase
		label.details = ""
		if(context instanceof TransitionAction || context instanceof GuardAction){
			label.details = context.eContainer.getTransitionDetails(flattener)
		}
		return label
	}

	/**
	 * Get details for identifying transition or guard since they require more informaiton
	 * in order to distinguish among other elements.
	 */
	static def String getTransitionDetails(EObject object, FlatteningTransformer flattener) {
			var container = object
			while (container != null) {
				if (container instanceof Transition) {
					var CommonElement source = container.sourceVertex
					if(flattener.isNewElement(source)){
						source = flattener.getOriginalOwner(source)
					}else{
						source = flattener.getOriginalElement(source)
					}
					if(source == null){
						CodeGenPlugin.debug("source is null : " + (source as NamedElement).name)
					}
					var CommonElement target = container.targetVertex
					if(flattener.isNewElement(target)){
						target = flattener.getOriginalOwner(target)
					}else{
						target = flattener.getOriginalElement(target)
					}	
					if(target == null){
						CodeGenPlugin.debug("target is null : " + (target as NamedElement).name)
					}
					if(target != null && source != null){
						var sourceQname = QualifiedNames.cachedFullSMName(source as NamedElement)
						var targetQname = QualifiedNames.cachedFullSMName(target as NamedElement)
						var details = new UserEditableRegion.TransitionDetails(sourceQname, targetQname)
						for (t : container.triggers) {
							val ports = new ArrayList
							for (p : (t as RTTrigger).ports) {
								ports.add(p.name)
							}
							details.addTriggerDetail((t as RTTrigger).signal.name, ports)
						}
						return details.tagString
					}else{
						CodeGenPlugin.error("Source vertex and target vertex should not be null")
						return ""
					}
				}
				container = container.eContainer
		}
		""
	}

	/** 
	 * Calculate qualified name for object's container in the context of UMLRT codegen
	 */
	static def String getQualifiedName(EObject eobject, EObject capsule) {
		var EObject container = eobject
		var result = QualifiedNames.cachedFullName(capsule as NamedElement)
		
		while (container != null && !(container.eContainer instanceof StateMachine)) {
			if (container instanceof State) {
				val smqname = QualifiedNames.cachedFullSMName(container as NamedElement)
				return result + QualifiedNames.SEPARATOR + smqname
			}

			container = container.eContainer
		}

		return result
	}

}
