/*******************************************************************************
 * Copyright (c) 2014-2016 Zeligsoft (2009) Limited and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/

package org.eclipse.papyrusrt.codegen.papyrus.cdt;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map.Entry;

import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.papyrus.infra.core.resource.ModelSet;
import org.eclipse.papyrusrt.codegen.CodeGenPlugin;
import org.eclipse.papyrusrt.codegen.UserEditableRegion;
import org.eclipse.papyrusrt.codegen.UserEditableRegion.Label;
import org.eclipse.papyrusrt.xtumlrt.util.QualifiedNames;
import org.eclipse.uml2.common.util.UML2Util;
import org.eclipse.uml2.uml.CallEvent;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Constraint;
import org.eclipse.uml2.uml.Event;
import org.eclipse.uml2.uml.OpaqueBehavior;
import org.eclipse.uml2.uml.OpaqueExpression;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.Port;
import org.eclipse.uml2.uml.Region;
import org.eclipse.uml2.uml.State;
import org.eclipse.uml2.uml.StateMachine;
import org.eclipse.uml2.uml.Transition;
import org.eclipse.uml2.uml.Trigger;
import org.eclipse.uml2.uml.UMLPackage;
import org.eclipse.uml2.uml.ValueSpecification;
import org.eclipse.uml2.uml.Vertex;

/**
 * 
 * @author ysroh
 *
 */
public class UMLEObjectLocator implements org.eclipse.papyrusrt.codegen.IEObjectLocator {


	/**
	 * Language.
	 */
	public static final String LANGUAGE = "C++";

	/**
	 * Model set.
	 */
	private ModelSet modelSet = new ModelSet();

	/**
	 * Constructor.
	 *
	 */
	public UMLEObjectLocator() {
	}

	/**
	 * Retrieve root element of given resource uri.
	 * 
	 * @param uri
	 *            resource uri
	 * @return root package
	 */
	private Package getRoot(String uri) {

		Package result = null;
		URI resourceUri = URI.createURI(uri);
		if ("uml".equals(resourceUri.fileExtension())) {
			Resource r = modelSet.getResource(resourceUri, true);
			result = (Package) EcoreUtil.getObjectByType(r.getContents(), UMLPackage.Literals.PACKAGE);
		}
		return result;
	}

	@Override
	public EObject getEObject(UserEditableRegion.Label label) {

		List<String> qualifiedNames = new ArrayList<>();
		qualifiedNames.addAll(Arrays.asList(label.getQualifiedName().split(QualifiedNames.SEPARATOR)));

		Package nextPackage = getRoot(label.getUri());
		if (nextPackage == null) {
			// resource not found
			return null;
		}
		Class capsuleOrClass = null;
		qualifiedNames.remove(0);
		while (!qualifiedNames.isEmpty()) {
			String qname = qualifiedNames.remove(0);
			Package pkg = nextPackage.getNestedPackage(qname);
			if (pkg == null) {
				capsuleOrClass = (Class) nextPackage.getPackagedElement(qname, false, UMLPackage.Literals.CLASS, false);
				break;
			}
			nextPackage = pkg;
		}

		EObject result = null;

		if (capsuleOrClass == null) {
			CodeGenPlugin.error("Element not found : " + label.getQualifiedName());
		} else {
			if (label.getType().equals(UMLPackage.Literals.OPERATION.getName().toLowerCase())) {
				Operation op = capsuleOrClass.getOwnedOperation(label.getDetails(), null, null);
				OpaqueBehavior method = null;
				if (op.getMethods().isEmpty()) {
					method = (OpaqueBehavior) capsuleOrClass.createOwnedBehavior(null, UMLPackage.Literals.OPAQUE_BEHAVIOR);
					op.getMethods().add(method);
				} else {
					method = (OpaqueBehavior) op.getMethods().get(0);
				}
				if (!method.getLanguages().contains(LANGUAGE)) {
					method.getLanguages().add(LANGUAGE);
				}
				result = method;
			} else {
				// This is element under statemachine.
				EObject smElement = getSMElement(capsuleOrClass, qualifiedNames);
				result = getUserCodeElement(smElement, label.getType(), label.getDetails());

			}
		}

		return result;
	}

	/**
	 * Queries the element that holds the user code.
	 * 
	 * @param container
	 *            cotainer
	 * @param type
	 *            element type
	 * @param details
	 *            element details
	 * @return element containing user code
	 */
	private EObject getUserCodeElement(EObject container, String type, String details) {
		EObject result = null;
		if (type.equals(UMLPackage.Literals.TRANSITION.getName().toLowerCase())) {
			Transition t = getTransition(container, details);
			if (t != null) {
				result = t.getEffect();
			}
		} else if (type.equals(UMLPackage.Literals.STATE__ENTRY.getName())) {
			if (container instanceof State) {
				State s = (State) container;
				result = s.getEntry();
			}
		} else if (type.equals(UMLPackage.Literals.STATE__EXIT.getName())) {
			if (container instanceof State) {
				State s = (State) container;
				result = s.getExit();
			}
		} else if (type.equals(UMLPackage.Literals.TRANSITION__GUARD.getName())) {
			Transition t = getTransition(container, details);
			if (t != null) {
				result = t.getGuard();
			}
		}
		return result;
	}

	/**
	 * Find element under State machine.
	 * 
	 * @param capsuleOrClass
	 *            Capsule or class containing state machine
	 * @param qnameRelativeToSM
	 *            qualified name
	 * @return EObject
	 */
	private EObject getSMElement(Class capsuleOrClass, List<String> qnameRelativeToSM) {
		StateMachine sm = (StateMachine) capsuleOrClass.getOwnedBehaviors().get(0);
		Region r = sm.getRegions().get(0);
		Region nextRegion = r;
		EObject result = sm;

		while (!qnameRelativeToSM.isEmpty()) {
			Vertex v = nextRegion.getSubvertex(qnameRelativeToSM.remove(0));
			if (v == null) {
				CodeGenPlugin.error("No matching element found.");
				break;
			} else if (!((State) v).getRegions().isEmpty()) {
				nextRegion = ((State) v).getRegions().get(0);
			}
			result = v;
		}

		if (!qnameRelativeToSM.isEmpty()) {
			// something wrong
			CodeGenPlugin.error("No matching element found");
			result = null;
		}

		return result;
	}

	/**
	 * Retrieve Transition from Capsule or Class.
	 * 
	 * @param container
	 *            State machine or State that contains this transition
	 * @param transitionDetails
	 *            Transition details
	 * @return Transition
	 */
	private Transition getTransition(EObject container, String transitionDetails) {
		Transition result = null;
		Region region = null;
		if (container instanceof StateMachine) {
			region = ((StateMachine) container).getRegions().get(0);
		} else {
			region = ((State) container).getRegions().get(0);
		}

		if (region != null) {
			UserEditableRegion.TransitionDetails details = new UserEditableRegion.TransitionDetails(transitionDetails);

			for (Transition t : region.getTransitions()) {
				if (getSMQualifiedName(t.getSource()).equals(details.getSourceQname()) && getSMQualifiedName(t.getTarget()).equals(details.getTargetQname())) {
					if (details.getTriggerDetailMap().isEmpty()) {
						result = t;
						break;
					}

					// check if the number of triggers are same. Go to next transition if different.
					if (details.getTriggerDetailMap().size() != t.getTriggers().size()) {
						continue;
					}

					boolean found = false;
					for (Entry<String, List<String>> entry : details.getTriggerDetailMap().entrySet()) {
						found = false;
						for (Trigger trigger : t.getTriggers()) {
							Event event = trigger.getEvent();
							if (event == null || !(event instanceof CallEvent)) {
								continue;
							}
							Operation message = ((CallEvent) event).getOperation();
							if (message != null && entry.getKey().equals(message.getName()) && comparetriggerPorts(trigger.getPorts(), entry.getValue())) {
								found = true;
								break;
							}
						}
						if (!found) {
							// if trigger details do not match with current transition then go to next one.
							break;
						}
					}
					if (found) {
						// if all triggers found then this is the transition we are looking for.
						result = t;
						break;
					}
				}
			}
		}

		return result;
	}

	/**
	 * Compare actual trigger ports with expected port names.
	 * 
	 * @param ports
	 *            actual ports
	 * @param expectedPorts
	 *            expected port names
	 * @return boolean
	 */
	private boolean comparetriggerPorts(List<Port> ports, List<String> expectedPorts) {

		List<String> actualPorts = new ArrayList<>();
		for (Port p : ports) {
			actualPorts.add(p.getName());
		}
		actualPorts.retainAll(expectedPorts);
		return actualPorts.size() == expectedPorts.size();

	}

	/**
	 * Calculate qualified name relative to State Machine.
	 * 
	 * @param eObject
	 *            EObject
	 * @return QualifiedName
	 */
	public static String getSMQualifiedName(EObject eObject) {
		String result = UML2Util.EMPTY_STRING;
		EObject container = eObject;
		while (container != null && !(container instanceof StateMachine)) {
			if (container instanceof Vertex) {
				if (result.length() != 0) {
					result = QualifiedNames.SEPARATOR + result;
				}
				result = ((Vertex) container).getName() + result;
			}
			container = container.eContainer();
		}
		return result;
	}

	/**
	 * @see org.eclipse.papyrusrt.codegen.IEObjectLocator#saveSource(org.eclipse.papyrusrt.codegen.UserEditableRegion.Label, java.lang.String)
	 *
	 * @param label
	 * @param source
	 * @return
	 */
	@Override
	public EObject saveSource(Label label, String source) {
		EObject eo = getEObject(label);
		if (eo instanceof OpaqueBehavior) {
			OpaqueBehavior ob = (OpaqueBehavior) eo;
			ob.getBodies().clear();
			ob.getBodies().add(source);
		} else if (eo instanceof Constraint) {
			Constraint guard = (Constraint) eo;
			ValueSpecification value = guard.getSpecification();
			if (value instanceof OpaqueExpression) {
				((OpaqueExpression) value).getBodies().clear();
				((OpaqueExpression) value).getBodies().add(source);
			}
		}
		return eo;
	}
}
