/*******************************************************************************
 * Copyright (c) 2014-2016 Zeligsoft (2009) Limited and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/

package org.eclipse.papyrusrt.codegen.papyrus.handlers;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.commands.AbstractHandler;
import org.eclipse.core.commands.ExecutionEvent;
import org.eclipse.core.commands.ExecutionException;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.runtime.MultiStatus;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.IStructuredSelection;
import org.eclipse.papyrus.infra.emf.utils.EMFHelper;
import org.eclipse.papyrusrt.codegen.UserEditableRegion;
import org.eclipse.papyrusrt.codegen.UserEditableRegion.Label;
import org.eclipse.papyrusrt.codegen.papyrus.Activator;
import org.eclipse.papyrusrt.codegen.papyrus.UMLRTGenerator;
import org.eclipse.papyrusrt.codegen.papyrus.cdt.EditorUtil;
import org.eclipse.papyrusrt.codegen.papyrus.cdt.UMLEObjectLocator;
import org.eclipse.papyrusrt.codegen.papyrus.internal.UMLRTCppCodeGen;
import org.eclipse.papyrusrt.umlrt.core.utils.StateUtils;
import org.eclipse.papyrusrt.xtumlrt.util.QualifiedNames;
import org.eclipse.ui.handlers.HandlerUtil;
import org.eclipse.uml2.uml.CallEvent;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Constraint;
import org.eclipse.uml2.uml.Element;
import org.eclipse.uml2.uml.Event;
import org.eclipse.uml2.uml.NamedElement;
import org.eclipse.uml2.uml.OpaqueBehavior;
import org.eclipse.uml2.uml.Operation;
import org.eclipse.uml2.uml.Port;
import org.eclipse.uml2.uml.State;
import org.eclipse.uml2.uml.StateMachine;
import org.eclipse.uml2.uml.Transition;
import org.eclipse.uml2.uml.Trigger;
import org.eclipse.uml2.uml.UMLPackage;

/**
 * @author ysroh
 *
 */
public class EditSourceActionHandler extends AbstractHandler {


	/**
	 * Constructor.
	 *
	 */
	public EditSourceActionHandler() {
	}

	/**
	 * @see org.eclipse.core.commands.IHandler#execute(org.eclipse.core.commands.ExecutionEvent)
	 *
	 * @param event
	 * @return
	 * @throws ExecutionException
	 */
	@Override
	public Object execute(ExecutionEvent event) throws ExecutionException {

		ISelection sel = HandlerUtil.getCurrentSelection(event);
		if (sel instanceof IStructuredSelection) {

			IStructuredSelection selection = (IStructuredSelection) sel;
			Object obj = selection.getFirstElement();
			final EObject eobj = EMFHelper.getEObject(obj);
			if (eobj == null) {
				throw new IllegalArgumentException(obj.getClass().getCanonicalName() + " is not an EObject");
			}

			EObject root = EcoreUtil.getRootContainer(eobj);
			List<EObject> targets = new ArrayList<>();
			targets.add(root);
			String top = UMLRTCppCodeGen.getTopCapsuleName((Element) root);
			UMLRTGenerator.getInstance().generate(targets, top, true);

			MultiStatus rc = new MultiStatus(Activator.PLUGIN_ID, 0, "Edit Source Action Status", null);

			if (eobj instanceof OpaqueBehavior) {
				// User selected actual opaque behaviour to edit source
				EObject container = eobj.eContainer();
				IFile file = getFile((NamedElement) container);
				if (container instanceof Transition) {
					// This is transition action
					Label label = getLabel(container);
					EditorUtil.openEditor(file, label, rc);
				} else if (container instanceof State) {
					// This is either entry or exit action
					Label label = getLabel(container);
					State s = (State) container;
					if (s.getEntry() == eobj) {
						label.setType(UMLPackage.Literals.STATE__ENTRY.getName());
					} else {
						label.setType(UMLPackage.Literals.STATE__EXIT.getName());
					}
					EditorUtil.openEditor(file, label, rc);
				}
			} else {
				Label label = getLabel(eobj);
				IFile file = getFile((NamedElement) eobj);
				EditorUtil.openEditor(file, label, rc);
			}
		}
		return null;
	}

	/**
	 * Compute label.
	 * 
	 * @param context
	 *            EObject
	 * @return label info
	 */
	public Label getLabel(EObject context) {
		EObject container = context.eContainer();
		Label label = new Label();
		label.setQualifiedName(getParentQualifiedName(context));
		if (context instanceof Transition) {
			label.setType(UMLPackage.Literals.TRANSITION.getName().toLowerCase());
			label.setDetails(getDetails((Transition) context));
		} else if (context instanceof Constraint) {
			label.setType(UMLPackage.Literals.TRANSITION__GUARD.getName().toLowerCase());
			label.setDetails(getDetails((Transition) container));
		} else if (context instanceof Operation) {
			label.setType(UMLPackage.Literals.OPERATION.getName().toLowerCase());
			label.setDetails(((Operation) context).getName());
		} else if (context instanceof State) {
			label.setQualifiedName(label.getQualifiedName() + QualifiedNames.SEPARATOR
					+ ((State) context).getName());
			// Entry or Exit?
			State s = (State) context;
			if (s.getEntry() != null) {
				label.setType(UMLPackage.Literals.STATE__ENTRY.getName());
			} else {
				label.setType(UMLPackage.Literals.STATE__EXIT.getName());
			}
		} else {
			label.setType(UMLPackage.Literals.CLASS.getName());
		}
		label.setUri(context.eResource().getURI().toString());

		return label;
	}

	/**
	 * Compute transition details.
	 * 
	 * @param transition
	 *            Transition.
	 * @return transition detail string
	 */
	public String getDetails(Transition transition) {
		String sourceQname = UMLEObjectLocator.getSMQualifiedName(transition.getSource());
		String targetQname = UMLEObjectLocator.getSMQualifiedName(transition.getTarget());
		UserEditableRegion.TransitionDetails details = new UserEditableRegion.TransitionDetails(sourceQname, targetQname);
		for (Trigger t : transition.getTriggers()) {
			List<String> ports = new ArrayList<>();
			for (Port p : t.getPorts()) {
				ports.add(p.getName());
			}
			Event event = t.getEvent();
			if (event != null && event instanceof CallEvent) {
				Operation message = ((CallEvent) event).getOperation();
				if (message != null && message.getName() != null) {
					details.addTriggerDetail(message.getName(), ports);
				}
			}
		}
		return details.getTagString();
	}

	/**
	 * Compute qualified name for container.
	 * 
	 * @param object
	 *            object to compute
	 * @return qualified name
	 */
	public String getParentQualifiedName(EObject object) {
		List<String> names = new ArrayList<>();
		EObject container = object.eContainer();
		while (container != null) {
			if (container instanceof State && StateUtils.isRTState((State) container)) {
				names.add(0, ((State) container).getName());
			} else if (container instanceof Class && !(container instanceof StateMachine)) {
				names.add(0, ((Class) container).getName());
			} else if (container instanceof org.eclipse.uml2.uml.Package) {
				names.add(0, ((org.eclipse.uml2.uml.Package) container).getName());
			}

			container = container.eContainer();
		}
		return String.join(QualifiedNames.SEPARATOR, names);
	}

	/**
	 * Get generated source file for given element.
	 * 
	 * @param element
	 *            Element.
	 * @return generated source file
	 */
	public IFile getFile(NamedElement element) {
		IProject project = UMLRTGenerator.getInstance().getProject(element);
		Class clazz = null;
		EObject container = element;
		while (container != null) {
			if (container instanceof Class && !(container instanceof StateMachine)) {
				clazz = (Class) container;
				break;
			}
			container = container.eContainer();
		}
		if (clazz != null) {
			IFolder srcFolder = project.getFolder("src");
			return srcFolder.getFile(clazz.getName() + ".cc");
		}
		return null;
	}

}
