/*******************************************************************************
 * Copyright (c) 2014-2015 Zeligsoft (2009) Limited and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package org.eclipse.papyrusrt.codegen.utils;

import org.eclipse.uml2.uml.Behavior;
import org.eclipse.uml2.uml.Constraint;
import org.eclipse.uml2.uml.OpaqueBehavior;
import org.eclipse.uml2.uml.OpaqueExpression;
import org.eclipse.uml2.uml.Operation;

/**
 * Utilities for C++ action code in UML2 models.
 * 
 * @author epp
 */
public final class UML2CppUtil {

	/**
	 * Default Constructor. Private as this is a utility class.
	 *
	 */
	private UML2CppUtil() {
	}

	/**
	 * Return true if the given string represents an OpaqueBehavior's C++ language
	 * code and false otherwise.
	 * 
	 * @param lang
	 *            - The language of an opaque behaviour.
	 * @return True if the given string represents an OpaqueBehavior's C++ language
	 *         code and false otherwise.
	 */
	public static boolean isCppLanguage(String lang) {
		return "C++".equals(lang)
				|| "C/C++".equals(lang)
				|| "C".equals(lang)
				|| "CPP".equals(lang);
	}

	/**
	 * Return the C++ code string associated with the given constraint. Returns null if
	 * there is no such code.
	 * 
	 * @param constraint
	 *            - A {@link Constraint}.
	 * @return The {@link String} with the opaque C/C++ behaviour.
	 */
	public static String getCppCode(Constraint constraint) {
		String code = null;
		if (constraint.getSpecification() instanceof OpaqueExpression) {
			OpaqueExpression opaqueExpression = (OpaqueExpression) constraint.getSpecification();
			for (int i = 0; i < opaqueExpression.getLanguages().size(); ++i) {
				String language = opaqueExpression.getLanguages().get(i);
				if (isCppLanguage(language)) {
					if (code != null) {
						throw new RuntimeException("too many C++-compatible languages for " + constraint.getQualifiedName());
					}
					code = opaqueExpression.getBodies().get(i);
				}
			}
		}

		return code;
	}

	/**
	 * Return the C++ code string associated with the given behaviour. Returns null if
	 * there is no such code.
	 * 
	 * @param behaviour
	 *            - A {@link Behavior}.
	 * @return The C++ code string associated with the given behaviour. Returns null if
	 *         there is no such code.
	 */
	public static String getCppCode(Behavior behaviour) {
		String code = null;
		if (behaviour instanceof OpaqueBehavior) {
			OpaqueBehavior opaqueBehaviour = (OpaqueBehavior) behaviour;
			for (int i = 0; i < opaqueBehaviour.getLanguages().size(); ++i) {
				String language = opaqueBehaviour.getLanguages().get(i);
				if (isCppLanguage(language)) {
					if (code != null) {
						throw new RuntimeException("too many C++-compatible languages for " + behaviour.getQualifiedName());
					}
					code = opaqueBehaviour.getBodies().get(i);
				}
			}
		}

		return code;
	}

	/**
	 * Return the C++ code string associated with the given operation. Returns null if
	 * there is no such code.
	 * 
	 * @param operation
	 *            - A {@link Operation}.
	 * @return The C++ code string associated with the given operation. Returns null if
	 *         there is no such code.
	 */
	public static String getCppCode(Operation operation) {
		String code = null;
		for (Behavior behaviour : operation.getMethods()) {
			String cppCode = getCppCode(behaviour);
			if (cppCode != null) {
				if (code == null) {
					code = cppCode;
				} else {
					throw new RuntimeException("too many C++-compatible behaviours for " + operation.getQualifiedName());
				}
			}
		}

		return code;
	}

}
