/*******************************************************************************
 * Copyright (c) 2014-2016 Zeligsoft (2009) Limited and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package org.eclipse.papyrusrt.codegen;

import java.util.Map;
import java.util.TreeMap;

import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtension;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.osgi.util.NLS;

/**
 * EObject locator.
 * 
 * @author ysroh
 *
 */
public final class EObjectLocator {
	/**
	 * Locator instance.
	 */
	private static EObjectLocator locator = new EObjectLocator();
	/**
	 * 
	 */
	private static final String EOBJECTLOCATOR_ELEMENT = "eobjectlocator";

	/**
	 * extension attribute.
	 */
	private static final String ID_ATTR = "id";
	/**
	 * extension attribute.
	 */
	private static final String CLASS_ATTR = "class";

	/**
	 * EObject locator registered through extension point.
	 */
	private final Map<String, IEObjectLocator> locators = new TreeMap<>();

	/**
	 * 
	 * Constructor.
	 *
	 */
	private EObjectLocator() {
		load();
	}

	public static EObjectLocator getInstance() {
		return locator;
	}

	/**
	 * Query EObject to registered EObject locator.
	 * 
	 * @param label
	 *            label info
	 * @return EObject
	 */
	public EObject getEObject(UserEditableRegion.Label label) {
		EObject result = null;
		URI uri = URI.createURI(label.getUri());
		IEObjectLocator locator = this.locators.get(uri.fileExtension().toLowerCase());
		if (locator != null) {
			result = locator.getEObject(label);
		}
		return result;
	}


	/**
	 * Query EObject to registered EObject locator.
	 * 
	 * @param label
	 *            label info
	 * @param source
	 *            source code to save
	 * @return EObject
	 */
	public EObject saveSource(UserEditableRegion.Label label, String source) {
		EObject result = null;
		URI uri = URI.createURI(label.getUri());
		IEObjectLocator locator = this.locators.get(uri.fileExtension().toLowerCase());
		if (locator != null) {
			result = locator.saveSource(label, source);
		}
		return result;
	}

	/**
	 * clear locators.
	 */
	public void dispose() {
		this.locators.clear();
	}

	/**
	 * Read extension point for locators.
	 */
	public void load() {
		IExtension[] extensions = Platform.getExtensionRegistry()
				.getExtensionPoint(CodeGenPlugin.ID, CodeGenPlugin.EOBJECTLOCATOR_EP_ID)
				.getExtensions();

		for (IExtension extension : extensions) {
			for (IConfigurationElement configElement : extension.getConfigurationElements()) {
				if (EOBJECTLOCATOR_ELEMENT.equals(configElement.getName())) {
					String id = configElement.getAttribute(ID_ATTR);
					if (id == null) {
						CodeGenPlugin.error(
								NLS.bind("Missing attribute: ", ID_ATTR));
						continue;
					}

					Object locatorObj = null;
					try {
						locatorObj = configElement.createExecutableExtension(CLASS_ATTR);
					} catch (CoreException e) {
						CodeGenPlugin.error(e);
					}
					if (locatorObj == null) {
						CodeGenPlugin.error(
								NLS.bind("Missing class: ", CLASS_ATTR));
						continue;
					}

					if (!(locatorObj instanceof IEObjectLocator)) {
						CodeGenPlugin.error(
								NLS.bind("Invalid class: ", CLASS_ATTR),
								null);
						continue;
					}

					this.locators.put(id, (IEObjectLocator) locatorObj);
				}
			}
		}
	}
}
