package org.eclipse.papyrusrt.codegen.cpp.test.codecompare;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.logging.Level;

import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.core.runtime.Status;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.emf.ecore.util.EcoreUtil;
import org.eclipse.papyrusrt.xtumlrt.util.PathUtils;
import org.eclipse.papyrusrt.xtumlrt.util.XTUMLRTLogger;
import org.eclipse.uml2.uml.UMLPackage;
import org.junit.Test;
import org.osgi.framework.Bundle;

/**
 * Unit test to compare the output of the code generator.
 */
public class CodeGenCompareTest {

	/** The plugin id. */
	public static final String PLUGIN_ID = "org.eclipse.papyrusrt.codegen.cpp.test";

	/** The resource set where the models will be loaded. */
	protected static ResourceSetImpl resourceSet = new ResourceSetImpl();

	/** The folder containing the models whose generated code will be compared. */
	private static final String MODEL_PATH = "models";

	/** The list of models to test. */
	private static final String[] TEST_MODELS = {
			"PingPong",
			// "PingPong-data",
			// "PingPong-dynamic",
			"ComputerSystem"
	};

	private static final List<String> TEST_MODEL_LIST = Arrays.asList(TEST_MODELS);

	/** The prefixes for lines that are to be ignored by the comparison. */
	private static final String[] PREFIXES_TO_IGNORE = {
			"/*",
			"# Generated",
			"UMLRTS_ROOT",
			"set(UMLRTS_ROOT"
	};

	/**
	 * Constructor.
	 */
	public CodeGenCompareTest() {
		XTUMLRTLogger.getLogger().setLevel(Level.FINEST);
		UMLRTTestGenerator.INSTANCE.setupExternalPackageManagement(resourceSet);
	}

	/**
	 * Load test model.
	 * 
	 * @param fullURI
	 *            - The {@link URI} of the model to load.
	 * @return The loaded UML {@link org.eclipse.uml2.uml.Package Package}.
	 */
	public static org.eclipse.uml2.uml.Package loadModel(URI fullURI) {
		Resource resource = resourceSet.getResource(fullURI, true);
		EList<EObject> contents = resource.getContents();
		org.eclipse.uml2.uml.Package pkg = (org.eclipse.uml2.uml.Package) EcoreUtil.getObjectByType(contents, UMLPackage.Literals.PACKAGE);
		return pkg;
	}

	/**
	 * The basic compate test. For each test model, loads the model, invokes the code generator,
	 * and iterates through each of the files in the original model's "expected_src" folder
	 * comparing them with the file of the same name in the generated "src" folder.
	 */
	@Test
	public void testCompare() {
		Bundle bundle = Platform.getBundle(CodeGenCompareTest.PLUGIN_ID);
		Path path = new Path(MODEL_PATH);
		URL url = FileLocator.find(bundle, path, null);
		File modelsDir = null;
		java.net.URI uri = null;
		try {
			uri = FileLocator.resolve(url).toURI();
		} catch (Exception e1) {
			assertTrue(e1.getMessage(), false);
		}
		assertNotNull(uri);
		modelsDir = new File(PathUtils.simplifyPath(uri.getRawPath()));

		// test all the models under models folder
		for (File modelFolder : modelsDir.listFiles()) {
			if (!TEST_MODEL_LIST.contains(modelFolder.getName())) {
				continue;
			}
			String modelName = modelFolder.getName() + ".uml";
			System.out.println("Testing " + modelName);
			File modelFile = new File(modelFolder, modelName);
			assertTrue("Model file " + modelName + " does not exist", modelFile.exists());
			org.eclipse.uml2.uml.Package model = loadModel(URI.createFileURI(modelFile.toPath().toString()));
			List<EObject> elements = new ArrayList<>();
			elements.add(model);
			IStatus status = UMLRTTestGenerator.INSTANCE.generate(elements, "Top", true);
			assertTrue("Code generation failed!", Status.ERROR != status.getSeverity());
			File expectedSrcFolder = new File(modelFolder, "expected_src");
			File outputFolder = UMLRTTestGenerator.INSTANCE.getOutputFolder(model);
			List<String> results = new ArrayList<>();
			for (File expected : expectedSrcFolder.listFiles()) {
				File generated = new File(outputFolder, expected.getName());
				assertTrue(generated.getAbsolutePath() + " does not exist", generated.exists());
				try {
					System.out.println("Comparing " + expected.getName());
					if (!compare(expected, generated)) {
						results.add(expected.getName());
					}
				} catch (Exception e) {
					assertEquals("Failed to compare text files", true, false);
				}
			}
			// ToDo: to be uncommented once the bug fixed.
			assertTrue(modelFolder.getName() + " does not match with expected src. Files: " + results.toString(),
					results.isEmpty());
		}
	}

	/**
	 * Compares two text files line by line.
	 * 
	 * @param expected
	 *            - A {@link File}.
	 * @param generated
	 *            - Another {@link File}.
	 * @return {@code true} iff the files are identical, other than for lines that are to be ignored. (see {@link #ignoreLine}).
	 * @throws IOException
	 *             If one of the files could not be opened or could not be read.
	 * @throws FileNotFound
	 *             If one of the files could not be found.
	 */
	private boolean compare(File expected, File generated) throws IOException {

		boolean result = true;
		BufferedReader src = new BufferedReader(new FileReader(expected));
		BufferedReader target = new BufferedReader(new FileReader(generated));

		String strLine1 = null;
		String strLine2 = null;
		int lineNo = 0;

		while ((strLine1 = src.readLine()) != null) {
			strLine2 = target.readLine();
			lineNo++;
			if (strLine2 == null) {
				break;
			}
			strLine1 = strLine1.trim();
			strLine2 = strLine2.trim();
			if (ignoreLine(strLine1, strLine2)) {
				continue;
			}
			if (!strLine1.trim().equals(strLine2.trim())) {
				System.out.println(">> line " + lineNo + " : \"" + strLine1 + "\" :: \"" + strLine2 + "\"");
				result = false;
				break;
			}
		}
		if (result && target.readLine() != null) {
			result = false;
		}
		src.close();
		target.close();
		return result;
	}

	/**
	 * @param strLine1
	 *            - A {@link String}.
	 * @param strLine2
	 *            - A {@link String}.
	 * @return {@code true} iff both lines begin with one of the strings in {@link #PREFIXES_TO_IGNORE}.
	 */
	private boolean ignoreLine(String strLine1, String strLine2) {
		boolean ignore = false;
		for (String prefix : PREFIXES_TO_IGNORE) {
			if (strLine1.startsWith(prefix) && strLine2.startsWith(prefix)) {
				ignore = true;
				break;
			}
		}
		return ignore;
	}

}
