/*******************************************************************************
 * Copyright (c) 2014-2015 Zeligsoft (2009) Limited and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *******************************************************************************/
package org.eclipse.papyrusrt.codegen.cpp.test.codecompare;

import java.io.File;
import java.net.URL;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.core.resources.IFolder;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.FileLocator;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Platform;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.papyrusrt.codegen.CodeGenPlugin;
import org.eclipse.papyrusrt.codegen.cpp.AbstractUMLRT2CppCodeGenerator;
import org.eclipse.papyrusrt.codegen.cpp.CppProjectGenerator;
import org.eclipse.papyrusrt.codegen.cpp.XTUMLRT2CppCodeGenerator;
import org.eclipse.papyrusrt.codegen.utils.ProjectUtils;
import org.eclipse.papyrusrt.xtumlrt.util.PathUtils;
import org.eclipse.uml2.uml.NamedElement;
import org.osgi.framework.Bundle;

@SuppressWarnings("nls")
public class UMLRTTestGenerator extends AbstractUMLRT2CppCodeGenerator {

	public static UMLRTTestGenerator INSTANCE = new UMLRTTestGenerator();

	private UMLRTTestGenerator() {
		super();
		setStandalone(false);
	}

	@Override
	public IStatus generate(List<EObject> elements, String top, boolean uml) {

		IStatus status = super.generate(elements, top, uml);

		Set<IProject> projects = new HashSet<>();
		for (EObject e : elements) {
			IProject p = getProject(e);
			if (p != null) {
				projects.add(p);
			}
		}

		// Refresh generated projects
		for (IProject p : projects) {
			try {
				p.refreshLocal(IResource.DEPTH_INFINITE, null);
			} catch (CoreException e) {
				// ignore
			}
		}

		return status;
	}

	@Override
	protected File getModelFolder(EObject eobj) {
		if (eobj == null)
			return null;

		String eobjName = ((NamedElement) eobj).getName();
		String modelName = eobjName;
		org.eclipse.emf.common.util.URI eobjURI = eobj.eResource().getURI();
		int numSegments = eobjURI.segmentCount();
		if (numSegments > 1) {
			// If there are two or more segments, the resource is in a folder, so the second to last segment is the folder name.
			modelName = eobjURI.segment(numSegments - 2);
		} else if (numSegments == 1) {
			// If there is exactly one segment, the segment is the file itself.
			String segment = eobjURI.segment(0);
			int extPosition = segment.lastIndexOf(".uml");
			if (extPosition > 0) {
				modelName = segment.substring(0, extPosition);
			}
		}

		Bundle bundle = Platform.getBundle(CodeGenCompareTest.PLUGIN_ID);
		Path path = new Path("models/" + modelName);
		URL url = FileLocator.find(bundle, path, null);
		try {
			java.net.URI uri = FileLocator.resolve(url).toURI();
			return new File(PathUtils.simplifyPath(uri.getRawPath()));
		} catch (Exception e) {
		}
		return null;
	}

	public File getOutputFolder(EObject eobj) {
		return getOutputFolder(eobj, null);
	}

	@Override
	protected File getOutputFolder(EObject eobj, XTUMLRT2CppCodeGenerator codeGen) {
		IProject project = getProject(eobj, codeGen);

		// Bug 109: Refresh the folder before checking if it exists.
		IFolder folder = project.getFolder("src");
		try {
			folder.refreshLocal(IResource.DEPTH_ZERO, null);
		} catch (CoreException e) {
			CodeGenPlugin.error("could not refresh output folder", e);
		}

		if (!folder.exists())
			try {
				folder.create(true, true, null);
			} catch (CoreException e) {
				CodeGenPlugin.error("could not create output folder", e);
			}

		return folder.getRawLocation().toFile();
	}

	public IProject getProject(EObject eobj) {
		return getProject(eobj, null);
	}

	public IProject getProject(EObject eobj, XTUMLRT2CppCodeGenerator codeGen) {
		String projectName = ProjectUtils.getProjectName(eobj);
		projectName = projectName + "_CDTProject";

		String rootPath = ResourcesPlugin.getWorkspace().getRoot().getLocation().toOSString();

		IProject project = ResourcesPlugin.getWorkspace().getRoot().getProject(projectName);

		if (!project.exists()) {
			resetResource(eobj, codeGen);

			boolean result = new CppProjectGenerator().generate(rootPath, projectName);
			if (!result) {
				CodeGenPlugin.error("Failed to create output project");
				return null;
			}

			try {
				IProjectDescription description = ResourcesPlugin.getWorkspace()
						.loadProjectDescription(new org.eclipse.core.runtime.Path(rootPath + "/" + projectName
								+ "/.project"));
				project = ResourcesPlugin.getWorkspace().getRoot().getProject(description.getName());
				project.create(description, null);
			} catch (CoreException e1) {
				CodeGenPlugin.error("Failed to create output project");
			}
		}

		if (!project.isOpen()) {
			try {
				project.open(new NullProgressMonitor());
			} catch (CoreException e) {
				CodeGenPlugin.error("Failed to open project");
			}
		}

		return project;
	}
}