/*****************************************************************************
 * Copyright (c) 2014 CEA LIST.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  CEA LIST - Initial API and implementation
 *****************************************************************************/
package org.eclipse.papyrusrt.umlrt.core.tests.utils;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertFalse;
import static org.junit.Assert.assertTrue;
import static org.junit.Assert.fail;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.papyrus.infra.types.core.registries.ElementTypeSetConfigurationRegistry;
import org.eclipse.papyrus.junit.framework.classification.tests.AbstractPapyrusTest;
import org.eclipse.papyrus.junit.utils.rules.ModelSetFixture;
import org.eclipse.papyrus.junit.utils.rules.PluginResource;
import org.eclipse.papyrusrt.umlrt.core.Activator;
import org.eclipse.papyrusrt.umlrt.core.utils.RTPortKindEnum;
import org.eclipse.papyrusrt.umlrt.core.utils.RTPortUtils;
import org.eclipse.papyrusrt.umlrt.profile.UMLRealTime.PortRegistrationType;
import org.eclipse.papyrusrt.umlrt.profile.UMLRealTime.RTPort;
import org.eclipse.swt.widgets.Display;
import org.eclipse.uml2.uml.Class;
import org.eclipse.uml2.uml.Connector;
import org.eclipse.uml2.uml.Port;
import org.eclipse.uml2.uml.UMLPackage;
import org.junit.Assert;
import org.junit.BeforeClass;
import org.junit.ClassRule;

/**
 * Abstract class for {@link RTPort}
 */
@PluginResource("resource/Utils.di") // For documentation, has to be repeated in each concrete class...
public abstract class AbstractRTPortTest extends AbstractPapyrusTest {
	protected static Class capsule;

	protected static Port capsule_externalBehavior;
	protected static Port capsule_internalBehavior;
	protected static Port capsule_relay;
	protected static Port capsule_sap;
	protected static Port capsule_spp;
	protected static Port capsule_legacySap;
	protected static Port capsule_legacySpp;

	protected static Class connectedCapsule;
	protected static Class connectedCapsule2;
	protected static Port connectedCapsule_externalBehavior;
	protected static Port connectedCapsule_internalBehavior;
	protected static Port connectedCapsule_relay;
	protected static Port connectedCapsule_externalBehaviorNotConnected;
	protected static Port connectedCapsule_internalBehaviorNotConnected;
	protected static Port connectedCapsule_relayNotConnected;
	protected static Port connectedCapsule2_externalBehavior;
	protected static Port connectedCapsule2_internalBehavior;
	protected static Port connectedCapsule2_relay;
	protected static Connector connectedCapsule_externalBehaviorConnector;
	protected static Connector connectedCapsule_internalBehaviorConnector;
	protected static Connector connectedCapsule_relayConnector;

	@ClassRule
	public final static ModelSetFixture modelSetFixture = new ModelSetFixture();

	@BeforeClass
	public static void initModelElements() {
		// capsules
		capsule = (Class) modelSetFixture.getModel().getMember("Capsule");
		Assert.assertNotNull(capsule);

		// ports
		capsule_externalBehavior = capsule.getOwnedPort("externalBehavior", null);
		Assert.assertNotNull(capsule_externalBehavior);
		capsule_internalBehavior = capsule.getOwnedPort("internalBehavior", null);
		Assert.assertNotNull(capsule_internalBehavior);
		capsule_relay = capsule.getOwnedPort("relay", null);
		Assert.assertNotNull(capsule_relay);
		capsule_sap = capsule.getOwnedPort("sap", null);
		Assert.assertNotNull(capsule_sap);
		capsule_spp = capsule.getOwnedPort("spp", null);
		Assert.assertNotNull(capsule_spp);
		Assert.assertNotNull(capsule_relay);
		capsule_legacySap = capsule.getOwnedPort("legacySap", null);
		Assert.assertNotNull(capsule_legacySap);
		capsule_legacySpp = capsule.getOwnedPort("legacySpp", null);
		Assert.assertNotNull(capsule_legacySpp);

		// connections tests - classes
		connectedCapsule = (Class) modelSetFixture.getModel().getMember("ConnectedCapsule");
		Assert.assertNotNull(connectedCapsule);
		connectedCapsule2 = (Class) modelSetFixture.getModel().getMember("ConnectedCapsule2");
		Assert.assertNotNull(connectedCapsule2);

		// connections tests - ports
		connectedCapsule_externalBehavior = connectedCapsule.getOwnedPort("externalBehavior", null);
		Assert.assertNotNull(connectedCapsule_externalBehavior);
		connectedCapsule_internalBehavior = connectedCapsule.getOwnedPort("internalBehavior", null);
		Assert.assertNotNull(connectedCapsule_internalBehavior);
		connectedCapsule_relay = connectedCapsule.getOwnedPort("relay", null);
		Assert.assertNotNull(connectedCapsule_relay);
		// not connected
		connectedCapsule_externalBehaviorNotConnected = connectedCapsule.getOwnedPort("externalBehaviorNotConnected", null);
		Assert.assertNotNull(connectedCapsule_externalBehaviorNotConnected);
		connectedCapsule_internalBehaviorNotConnected = connectedCapsule.getOwnedPort("internalBehaviorNotConnected", null);
		Assert.assertNotNull(connectedCapsule_internalBehaviorNotConnected);
		connectedCapsule_relayNotConnected = connectedCapsule.getOwnedPort("relayNotConnected", null);
		Assert.assertNotNull(connectedCapsule_relayNotConnected);

		// port on part
		connectedCapsule2_externalBehavior = connectedCapsule2.getOwnedPort("externalBehavior", null);
		Assert.assertNotNull(connectedCapsule2_externalBehavior);
		connectedCapsule2_internalBehavior = connectedCapsule2.getOwnedPort("internalBehavior", null);
		Assert.assertNotNull(connectedCapsule2_internalBehavior);
		connectedCapsule2_relay = connectedCapsule2.getOwnedPort("relay", null);
		Assert.assertNotNull(connectedCapsule2_relay);

		// connector
		connectedCapsule_externalBehaviorConnector = connectedCapsule.getOwnedConnector("externalBehaviorConnector");
		Assert.assertNotNull(connectedCapsule_externalBehaviorConnector);
		connectedCapsule_internalBehaviorConnector = connectedCapsule.getOwnedConnector("internalBehaviorConnector");
		Assert.assertNotNull(connectedCapsule_internalBehaviorConnector);
		connectedCapsule_relayConnector = connectedCapsule.getOwnedConnector("relayConnector");
		Assert.assertNotNull(connectedCapsule_relayConnector);

		// ensure ElementTypeRegistry is started
		Display.getDefault().syncExec(new Runnable() {

			@Override
			public void run() {
				// force loading of the state machine plugin, but in UI thread, as the element types require some UI...
				ElementTypeSetConfigurationRegistry registry = ElementTypeSetConfigurationRegistry.getInstance();
				Activator.log.debug("Configuration size: " + registry.getElementTypeSetConfigurations().size());
			}
		});
	}

	protected void checkPort(EObject objectToSet, RTPortKindEnum kind) {
		Port port = null;
		if (objectToSet instanceof Port) {
			port = (Port) objectToSet;
		} else if (objectToSet instanceof RTPort) {
			port = ((RTPort) objectToSet).getBase_Port();
		} else {
			fail("impossible to get the port from EObject: " + objectToSet);
		}
		// See Bug 479352
		Assert.assertFalse("RTPort Multiplicities should be left unset ", port.eIsSet(UMLPackage.eINSTANCE.getMultiplicityElement().getEStructuralFeature("lowerValue")));
		Assert.assertFalse("RTPort Multiplicities should be left unset ", port.eIsSet(UMLPackage.eINSTANCE.getMultiplicityElement().getEStructuralFeature("upperValue")));

		switch (kind) {
		case EXTERNAL:
			assertTrue("Service: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isService(port));
			assertTrue("Wired: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isWired(port));
			assertTrue("Behavior: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isBehavior(port));
			assertFalse("Publish: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isPublish(port));
			assertEquals("Registration: error in RTPortUtils#getChangeKindCommand(Port)", PortRegistrationType.AUTOMATIC, RTPortUtils.getRegistration(port));
			assertEquals("RegistrationOverride: error in RTPortUtils#getChangeKindCommand(Port)", "", RTPortUtils.getRegistrationOverride(port));
			break;
		case INTERNAL:
			assertFalse("Service: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isService(port));
			assertTrue("Wired: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isWired(port));
			assertTrue("Behavior: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isBehavior(port));
			assertFalse("Publish: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isPublish(port));
			assertEquals("Registration: error in RTPortUtils#getChangeKindCommand(Port)", PortRegistrationType.AUTOMATIC, RTPortUtils.getRegistration(port));
			assertEquals("RegistrationOverride: error in RTPortUtils#getChangeKindCommand(Port)", "", RTPortUtils.getRegistrationOverride(port));
			break;
		case RELAY:
			assertTrue("Service: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isService(port));
			assertTrue("Wired: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isWired(port));
			assertFalse("Behavior: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isBehavior(port));
			assertFalse("Publish: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isPublish(port));
			assertEquals("Registration: error in RTPortUtils#getChangeKindCommand(Port)", PortRegistrationType.AUTOMATIC, RTPortUtils.getRegistration(port));
			assertEquals("RegistrationOverride: error in RTPortUtils#getChangeKindCommand(Port)", "", RTPortUtils.getRegistrationOverride(port));
			break;
		case SPP:
			// assertTrue("Service: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isService(port)); // legacy
			assertFalse("Wired: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isWired(port));
			assertTrue("Behavior: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isBehavior(port));
			assertTrue("Publish: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isPublish(port));
			break;
		case SAP:
			// assertTrue("Service: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isService(port));// legacy
			assertFalse("Wired: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isWired(port));
			assertTrue("Behavior: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isBehavior(port));
			assertFalse("Publish: error in RTPortUtils#getChangeKindCommand(Port)", RTPortUtils.isPublish(port));
			break;

		default:
			break;
		}
	}
}
