/*****************************************************************************
 * Copyright (c) 2016 CEA LIST.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *  CEA LIST - Initial API and implementation
 *****************************************************************************/
package org.eclipse.papyrusrt.umlrt.cpp.tests.lifecycle;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotNull;
import static org.junit.Assert.assertTrue;

import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.papyrus.infra.emf.utils.ServiceUtilsForEObject;
import org.eclipse.papyrus.junit.framework.classification.FailingTest;
import org.eclipse.papyrusrt.umlrt.core.defaultlanguage.IDefaultLanguage;
import org.eclipse.papyrusrt.umlrt.core.defaultlanguage.IDefaultLanguageService;
import org.eclipse.papyrusrt.umlrt.core.internal.defaultlanguage.NoDefautLanguage;
import org.eclipse.papyrusrt.umlrt.cpp.tests.AbstractPapyrusRTCppLibraryTest;
import org.eclipse.uml2.uml.Package;
import org.eclipse.uml2.uml.Profile;
import org.junit.Test;

/**
 * Tests on loading / unloading model with Cpp language
 */
public class CppModelLifecycleTest extends AbstractPapyrusRTCppLibraryTest {

	private static final int NUMBER_OF_CPP_PRIMITIVE_TYPES = 22;
	private static final int NUMBER_OF_SYSTEM_PROTOCOLS = 3;
	private static final int NUMBER_OF_SYSTEM_CLASSES = 4;
	private static final String UMLRT_CPP_ID = "umlrt-cpp";

	@Test
	public void testNoLanguage() throws Exception {
		IDefaultLanguageService service = ServiceUtilsForEObject.getInstance().getService(IDefaultLanguageService.class, cppRootModel);
		assertNotNull("DefaultLanguage service should not be null", service);
		IDefaultLanguage language = service.getActiveDefaultLanguage(noCppRootModel);
		assertEquals("DefaultLanguage should be noDefault", NoDefautLanguage.INSTANCE.getId(), language.getId());
	}

	@Test
	public void testLoadCppModel_LoadedLibraries() throws Exception {
		IDefaultLanguageService service = ServiceUtilsForEObject.getInstance().getService(IDefaultLanguageService.class, cppRootModel);
		assertNotNull("DefaultLanguage service should not be null", service);
		IDefaultLanguage language = service.getActiveDefaultLanguage(cppRootModel);
		assertEquals("DefaultLanguage should be cpp", UMLRT_CPP_ID, language.getId());

		// retrieve all cpp primitive types
		Resource cppResource = null;
		for (Resource resource : cppRootModel.eResource().getResourceSet().getResources()) {
			if (CPP_LIBRARY_URI.equals(resource.getURI())) {
				cppResource = resource;
			}
		}

		assertNotNull("Impossible to find Cpp library from cppRootModel", cppResource);
		assertEquals("Cpp Library should not be empty", 2, cppResource.getContents().size()); // package and stereotype "ModelLibrary"
		assertTrue("root element should be a Package", cppResource.getContents().get(0) instanceof Package);
		Package cpplibrary = (Package) cppResource.getContents().get(0);
		assertEquals("root element should be named AnsiCLibrary", "AnsiCLibrary", cpplibrary.getName());
		assertEquals("root element should have Types", NUMBER_OF_CPP_PRIMITIVE_TYPES, cpplibrary.getOwnedTypes().size());
	}

	@Test
	@FailingTest("Default Language Profile not created yet. based on a eannotation for now")
	public void testLoadCppModel_AppliedProfiles() throws Exception {
		Profile profile = cppRootModel.getAppliedProfile("DefaultLanguage");
		assertNotNull("Default Language profile is not applied", profile);
	}

	@Test
	public void testLoadCppModel_AccessibleTypes() throws Exception {
		IDefaultLanguageService service = ServiceUtilsForEObject.getInstance().getService(IDefaultLanguageService.class, cppRootModel);
		assertNotNull("DefaultLanguage service should not be null", service);
		IDefaultLanguage language = service.getActiveDefaultLanguage(cppRootModel);
		assertEquals("DefaultLanguage should be cpp", UMLRT_CPP_ID, language.getId());
		assertEquals("Unexpected number of types in the cpp library", NUMBER_OF_CPP_PRIMITIVE_TYPES, language.getSpecificPrimitiveTypes(cppRootModel.eResource().getResourceSet()).size());
	}


	@Test
	public void testLoadCppModel_SystemProtocols() throws Exception {
		IDefaultLanguageService service = ServiceUtilsForEObject.getInstance().getService(IDefaultLanguageService.class, cppRootModel);
		assertNotNull("DefaultLanguage service should not be null", service);
		IDefaultLanguage language = service.getActiveDefaultLanguage(cppRootModel);
		assertEquals("DefaultLanguage should be cpp", UMLRT_CPP_ID, language.getId());
		assertEquals("Unexpected number of system protocols in the cpp library", NUMBER_OF_SYSTEM_PROTOCOLS, language.getSystemProtocols(cppRootModel.eResource().getResourceSet()).size());
	}

	@Test
	public void testLoadCppModel_BaseProtocol() throws Exception {
		IDefaultLanguageService service = ServiceUtilsForEObject.getInstance().getService(IDefaultLanguageService.class, cppRootModel);
		assertNotNull("DefaultLanguage service should not be null", service);
		IDefaultLanguage language = service.getActiveDefaultLanguage(cppRootModel);
		assertEquals("DefaultLanguage should be cpp", UMLRT_CPP_ID, language.getId());
		assertNotNull("Base protocol should not be null in the cpp library", language.getBaseProtocol(cppRootModel.eResource().getResourceSet()));
		assertTrue("Base protocol should be abstract in the cpp library", language.getBaseProtocol(cppRootModel.eResource().getResourceSet()).isAbstract());
	}

	@Test
	public void testLoadCppModel_SystemClasses() throws Exception {
		IDefaultLanguageService service = ServiceUtilsForEObject.getInstance().getService(IDefaultLanguageService.class, cppRootModel);
		assertNotNull("DefaultLanguage service should not be null", service);
		IDefaultLanguage language = service.getActiveDefaultLanguage(cppRootModel);
		assertEquals("DefaultLanguage should be cpp", UMLRT_CPP_ID, language.getId());
		assertEquals("Unexpected number of system protocols in the cpp library", NUMBER_OF_SYSTEM_CLASSES, language.getSystemClasses(cppRootModel.eResource().getResourceSet()).size());
	}

}
